import numpy as np
import pandas as pd
import torch


def sample_data(n, n_env, invariance="T", post_treatment="collider", dim=10, seed=1331):
    np.random.seed(seed)
    data = []
    beta_xt = np.random.normal(size=(dim))
    beta_xt /= np.linalg.norm(beta_xt)

    beta_y = np.random.normal(size=dim)
    beta_y /= np.linalg.norm(beta_y)
    ate = 1
    print(f"beta_y: {beta_y}")
    print(f"ate: {ate}")
    for e in range(n_env):
        shift = np.random.normal(size=2)
        shift_x = np.random.normal(size=dim)
        X2 = np.random.normal(size=(n, dim), loc=shift_x, scale=shift_x**2)
        logit = X2 @ beta_xt
        if invariance == "T":
            Y = np.random.normal(size=n, loc=shift[0], scale=shift[0] ** 2)
            logit += np.random.normal(size=n)
        elif invariance == "Y":
            Y = np.random.normal(size=n)
            logit += np.random.normal(size=n, loc=shift[0], scale=shift[0] ** 2)
        else:
            Y = np.random.normal(size=n)
            logit += np.random.normal(size=n)

        p = torch.sigmoid(torch.Tensor(logit))
        T = torch.bernoulli(p).numpy()

        print(f"mean T: {T.mean()}")
        print(f"mean Y | T=1 : {Y[T==1].mean()}")
        print(f"mean Y | T=0 : {Y[T==0].mean()}")
        Y += X2 @ beta_y + ate * T

        if post_treatment == "collider":
            X1 = (
                (Y).reshape(n, 1)
                + T.reshape(n, 1)
                + np.random.normal(size=(n, 1), loc=shift[1], scale=shift[1] ** 2)
            )
        elif post_treatment == "noise":
            X1 = np.random.normal(size=(n, 1), loc=shift[1], scale=shift[1] ** 2)
        elif post_treatment == "descendant":
            X1 = (Y).reshape(-1, 1) + np.random.normal(
                size=(n, 1), loc=shift[1], scale=shift[1] ** 2
            )
        else:
            return

        X = np.hstack((X1, X2))
        # plot_data_classif(X2,T)

        environment_data = {
            "X": X,
            "X_inv": X2,
            "X_spu": X1,
            "T": T,
            "Y": Y,
            "ATE": ate,
        }
        data.append(environment_data)

    return data


def sample_semisynthetic(
    n_env,
    n_features,
    post_type="collider",
    n_post_treatment=1,
    n=None,
    ate=10,
    complexity="easy",
    invariance="T",
    mediator=False,
    seed=1331,
):
    np.random.seed(seed)
    data = []
    x = pd.read_csv("datasets/ihdp_obs.csv", delim_whitespace=True)
    X_columns = [f"X{i+1}" for i in range(n_features)]  # Use 6 features
    X_data = x[X_columns].values
    if n is None:
        n = len(X_data)

    dim = X_data.shape[1]  # Should be 6

    # Ensure at least 4 confounders
    effect_assignment = [None] * dim
    confounders = np.random.choice(range(dim), size=4, replace=False)
    for i in confounders:
        effect_assignment[i] = "both"

    # Assign the remaining features to affect T or Y
    remaining_indices = [i for i in range(dim) if i not in confounders]
    for i in remaining_indices:
        effect_assignment[i] = np.random.choice(["Y", "T"])

    # Identify features affecting Y and T
    features_affecting_Y = [
        i for i, effect in enumerate(effect_assignment) if effect in ["Y", "both"]
    ]
    features_affecting_T = [
        i for i, effect in enumerate(effect_assignment) if effect in ["T", "both"]
    ]

    print(f"Features affecting Y: {features_affecting_Y}")
    print(f"Features affecting T: {features_affecting_T}")

    # Generate beta coefficients or define transformations based on complexity
    if complexity == "easy":
        beta_y = np.zeros(dim)
        beta_t = np.zeros(dim)
        if features_affecting_Y:
            beta_y[features_affecting_Y] = np.random.normal(
                size=len(features_affecting_Y)
            )
            beta_y /= np.linalg.norm(beta_y)
        if features_affecting_T:
            beta_t[features_affecting_T] = np.random.normal(
                size=len(features_affecting_T)
            )
            beta_t /= np.linalg.norm(beta_t)
    elif complexity == "difficult":
        transformations_y = {}
        transformations_t = {}
        beta_y = {}
        beta_t = {}
        for i in range(dim):
            if effect_assignment[i] in ["Y", "both"]:
                transformations_y[i] = np.random.choice(
                    [
                        lambda x: 2.0 * np.log(np.abs(x)),
                        lambda x: (x / 2) ** 2,
                        lambda x: x + 1,
                        lambda x: np.exp(x / 2),
                    ]
                )
                beta_y[i] = np.random.uniform(-2, 2)
            if effect_assignment[i] in ["T", "both"]:
                transformations_t[i] = np.random.choice(
                    [
                        lambda x: 0.5 * np.log(np.abs(x) + 1),
                        lambda x: (x / 2) ** 2,
                        lambda x: x + 0.2,
                        lambda x: np.exp(x / 2),
                    ]
                )
                beta_t[i] = np.random.uniform(-0.5, 0.5)
    else:
        raise ValueError(
            f"Invalid complexity '{complexity}'. Choose 'easy' or 'difficult'."
        )

    # Identify features to hide based on invariance
    parents_of_Y = {i for i, effect in enumerate(effect_assignment) if effect == "Y"}
    parents_of_T = {i for i, effect in enumerate(effect_assignment) if effect == "T"}

    if invariance == "Y":
        hideable_features = list(parents_of_T)
    elif invariance == "T":
        hideable_features = list(parents_of_Y)
    else:
        hideable_features = list(parents_of_T | parents_of_Y)

    # # Select one feature to hide
    if len(hideable_features) >= 1:
        hidden_feature = np.random.choice(hideable_features, size=1, replace=False)
    else:
        hidden_feature = []

    # Coefficients for shifts
    coef_inv = np.random.uniform(0.5, 1.0)
    coef_x = np.random.uniform(0.5, 1.0)

    for e in range(n_env):
        # Create the shifts with sampled coefficients
        shift_inv = e * coef_inv + np.random.normal(size=1)
        shift_post = e * coef_x + np.random.normal(size=n_post_treatment)
        shift_x = e * coef_x + np.random.normal(size=dim)

        # Generate covariates with environment-specific shifts
        X2 = X_data[:n] + np.random.normal(loc=shift_x, scale=1.0, size=(n, dim))

        # Compute the logit for T and Y using the selected features
        if complexity == "easy":
            logit = X2 @ beta_t if features_affecting_T else np.zeros(n)
            Y = X2 @ beta_y if features_affecting_Y else np.zeros(n)
        elif complexity == "difficult":
            transformed_features_y = []
            transformed_features_t = []
            # For Y
            for i in transformations_y.keys():
                xi = X2[:, i]
                transformed_yi = transformations_y[i](xi)
                transformed_features_y.append(beta_y[i] * transformed_yi)

            # For T
            for i in transformations_t.keys():
                xi = X2[:, i]
                transformed_ti = transformations_t[i](xi)
                transformed_features_t.append(beta_t[i] * transformed_ti)
            Y = (
                np.sum(transformed_features_y, axis=0)
                if transformed_features_y
                else np.zeros(n)
            )
            logit = (
                np.sum(transformed_features_t, axis=0)
                if transformed_features_t
                else np.zeros(n)
            )

        # Apply invariance adjustments
        if invariance == "T":
            Y += np.random.normal(loc=shift_inv, scale=1.0, size=n)
            logit += np.random.normal(size=n)
        elif invariance == "Y":
            logit += np.random.normal(loc=shift_inv / 6, scale=1.0, size=n)
            Y += np.random.normal(size=n)
        elif invariance == "both":
            Y += np.random.normal(size=n)
            logit += np.random.normal(size=n)
        else:
            Y += np.random.normal(loc=shift_inv, scale=1.0, size=n)
            logit += np.random.normal(loc=shift_inv / 6, scale=1.0, size=n)

        # Compute treatment assignment and adjust the outcome
        p = torch.sigmoid(torch.Tensor(logit))
        T = torch.bernoulli(p).numpy()
        Y += ate * T

        # If mediator is True, create a mediator feature
        if mediator:
            mediator_beta = np.random.uniform(-1.0, 1.0)
            mediator_feature = mediator_beta * T + np.random.normal(size=n)
            Y += mediator_feature
            X2 = np.hstack((X2, mediator_feature.reshape(-1, 1)))  # Add mediator to X

        # Print treatment and outcome statistics for diagnostics
        print(f"mean T: {T.mean()}")
        print(f"mean Y | T=1 : {Y[T==1].mean()}")
        print(f"mean Y | T=0 : {Y[T==0].mean()}")

        # Create post-treatment variables
        post_treatment_variables = []
        for i in range(n_post_treatment):
            pt_type = post_type

            if pt_type == "collider":
                X1 = (
                    Y.reshape(n, 1)
                    + T.reshape(n, 1)
                    + np.random.normal(loc=shift_post[i], scale=1.0, size=(n, 1))
                )
            elif pt_type == "descendant":
                X1 = Y.reshape(-1, 1) + np.random.normal(
                    loc=shift_post[i], scale=1.0, size=(n, 1)
                )
            elif pt_type == "noise":
                X1 = np.random.normal(loc=shift_post[i], scale=1.0, size=(n, 1))

            post_treatment_variables.append(X1)

        # Stack all post-treatment variables together
        post_treatment_variables = np.hstack(post_treatment_variables)
        X2 = np.delete(X2, hidden_feature, axis=1)
        X = np.hstack((post_treatment_variables, X2))

        environment_data = {
            "X": X,
            "post": [i for i in range(n_post_treatment)],
            "confounders": [conf + n_post_treatment for conf in confounders],
            "hidden": (
                [hidden_feature[0] + n_post_treatment]
                if len(hidden_feature) != 0
                else []
            ),
            "T": T,
            "Y": Y,
            "ATE": ate,
            "effect_assignment": ["post" for i in range(n_post_treatment)]
            + effect_assignment,
        }
        data.append(environment_data)
    return data
