import argparse
import itertools

from data import sample_semisynthetic
from models.Kernelized import Kernelized
from models.KernelizedGumbel import KernelizedGumbel
from utils import (
    evaluate_baselines,
    evaluate_irm,
    evaluate_irm_nonlinear,
    evaluate_subset,
    pool_data,
)

param_grid = {
    "learning_rate": [0.01, 0.1, 0.001],
    "init_temp": [0.5, 0.8, 1.0],
    "anneal_rate": [0.9, 0.95, 0.99],
}

# Generate all combinations of parameters
param_combinations = list(
    itertools.product(
        param_grid["learning_rate"], param_grid["init_temp"], param_grid["anneal_rate"]
    )
)


def tune_kiv_gumbel(n_features, n_env, X, Y, T, use_xgboost):
    # Store results
    results = []

    # Sweep over randomly selected parameter combinations
    for lr, init_temp, anneal_rate in param_combinations:
        print(f"Testing with lr={lr}, init_temp={init_temp}, anneal_rate={anneal_rate}")

        # Initialize the model with current parameters
        kiv_gumbel = KernelizedGumbel(
            input_dim=n_features,
            n_env=n_env,
            learning_rate=lr,
            init_temp=init_temp,
            anneal_rate=anneal_rate,
            use_xgboost=use_xgboost,
        )

        # Compute the subset and loss
        subset_kiv_gumbel, loss_kiv_gumbel = kiv_gumbel.compute_subset(X, Y, T)

        # Store the results
        results.append(
            {
                "learning_rate": lr,
                "init_temp": init_temp,
                "anneal_rate": anneal_rate,
                "loss": loss_kiv_gumbel,
                "subset": subset_kiv_gumbel,
            }
        )

    # Find the best parameter combination
    best_result = min(results, key=lambda x: x["loss"])
    best_subset = best_result["subset"]
    print("Best parameters:")
    print(f"Learning Rate: {best_result['learning_rate']}")
    print(f"Initial Temperature: {best_result['init_temp']}")
    print(f"Anneal Rate: {best_result['anneal_rate']}")
    print(f"Lowest Loss: {best_result['loss']}")
    return best_subset


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--n_env",
        type=int,
        default=5,
        help="Number of environments",
    )
    parser.add_argument(
        "--mediator",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--n_features",
        type=int,
        default=6,
        help="Number of features",
    )
    parser.add_argument(
        "--n_post",
        type=int,
        default=1,
        help="Number of post-treatment features",
    )
    parser.add_argument(
        "--ate",
        type=float,
        default=2.0,
    )
    parser.add_argument(
        "--post_treatment",
        type=str,
        default="collider",
    )
    parser.add_argument(
        "--invariance",
        type=str,
        default="Y",
    )
    parser.add_argument(
        "--complexity",
        type=str,
        default="easy",
    )
    parser.add_argument(
        "--wandb_name",
        type=str,
        default="semi_main",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
    )
    parser.add_argument(
        "--gumbel_lr",
        type=float,
        default=0.01,
    )
    parser.add_argument(
        "--gumbel_init_temp",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--gumbel_anneal_rate",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--use_xgboost",
        action="store_true",
        default=False,
    )

    # Parse arguments
    args = parser.parse_args()

    data = sample_semisynthetic(
        mediator=args.mediator,
        n_env=args.n_env,
        n_features=args.n_features,
        post_type=args.post_treatment,
        n_post_treatment=args.n_post,
        ate=args.ate,
        complexity=args.complexity,
        invariance=args.invariance,
        seed=args.seed,
    )
    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")

    obs_features = X.shape[-1]

    if args.complexity == "difficult":
        error_irm = evaluate_irm_nonlinear(data)
    else:
        error_irm = evaluate_irm(data)

    error_all, error_null = evaluate_baselines(data, use_xgboost=args.use_xgboost)

    print(
        f"Baselines errors --  IRM: {error_irm}, All: {error_all}, Null: {error_null}"
    )

    subset_kiv_gumbel = tune_kiv_gumbel(
        obs_features, args.n_env, X, Y, T, args.use_xgboost
    )
    error_kiv_gumbel = evaluate_subset(
        subset_kiv_gumbel, data, use_xgboost=args.use_xgboost
    )
    print(f"Subset KIV gumbel: {subset_kiv_gumbel}, Error: {error_kiv_gumbel}")

    kiv = Kernelized(
        obs_features,
        args.n_env,
        linear_time=args.linear_time,
        use_xgboost=args.use_xgboost,
    )
    subset_kiv = kiv.compute_subset(X, Y, T)
    error_kiv = evaluate_subset(subset_kiv, data, use_xgboost=args.use_xgboost)

    print(f"Subset KIV: {subset_kiv}, Error: {error_kiv}")

    print(
        f"MAE (null):{error_null:.3f}\nMAE (all):{error_all:.3f}\nMAE(kiv):{error_kiv:.3f}\nMAE (kiv_gumbel):{error_kiv_gumbel:.3f}\nMAE (irm):{error_irm:.3f}"
    )


if __name__ == "__main__":
    main()
