import argparse
import itertools

from data import sample_data
from models.Kernelized import Kernelized
from models.KernelizedGumbel import KernelizedGumbel
from utils import evaluate_baselines, evaluate_irm, evaluate_subset, pool_data

param_grid = {
    "learning_rate": [0.01, 0.1, 0.001],
    "init_temp": [0.5, 0.8, 1.0],
    "anneal_rate": [0.9, 0.95, 0.99],
}


# Generate all combinations of parameters
param_combinations = list(
    itertools.product(
        param_grid["learning_rate"], param_grid["init_temp"], param_grid["anneal_rate"]
    )
)


def tune_kiv_gumbel(args, X, Y, T):
    # Store results
    results = []

    # Sweep over all parameter combinations
    for lr, init_temp, anneal_rate in param_combinations:
        print(f"Testing with lr={lr}, init_temp={init_temp}, anneal_rate={anneal_rate}")

        # Initialize the model with current parameters
        kiv_gumbel = KernelizedGumbel(
            input_dim=args.dim + 1,
            n_env=args.n_env,
            learning_rate=lr,
            init_temp=init_temp,
            anneal_rate=anneal_rate,
        )

        # Compute the subset and loss
        subset_kiv_gumbel, loss_kiv_gumbel = kiv_gumbel.compute_subset(X, Y, T)

        # Store the results
        results.append(
            {
                "learning_rate": lr,
                "init_temp": init_temp,
                "anneal_rate": anneal_rate,
                "loss": loss_kiv_gumbel,
                "subset": subset_kiv_gumbel,
            }
        )

    # Find the best parameter combination
    best_result = min(results, key=lambda x: x["loss"])
    best_subset = best_result["subset"]
    print("Best parameters:")
    print(f"Learning Rate: {best_result['learning_rate']}")
    print(f"Initial Temperature: {best_result['init_temp']}")
    print(f"Anneal Rate: {best_result['anneal_rate']}")
    print(f"Lowest Loss: {best_result['loss']}")
    return best_subset


def main():
    parser = argparse.ArgumentParser(description="Hi")
    parser.add_argument(
        "--n",
        type=int,
        default=1000,
        help="Number of samples for each env",
    )
    parser.add_argument(
        "--n_env",
        type=int,
        default=5,
        help="Number of environments",
    )
    parser.add_argument(
        "--dim",
        type=int,
        default=2,
        help="Dimension of X",
    )
    parser.add_argument(
        "--post_treatment",
        type=str,
        default="collider",
    )
    parser.add_argument(
        "--invariance",
        type=str,
        default="Y",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
    )
    parser.add_argument(
        "--gumbel_lr",
        type=float,
        default=0.01,
    )
    parser.add_argument(
        "--gumbel_init_temp",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--gumbel_anneal_rate",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--linear_time",
        type=bool,
        default=False,
    )

    # Parse arguments
    args = parser.parse_args()

    data = sample_data(
        n=args.n,
        n_env=args.n_env,
        invariance=args.invariance,
        post_treatment=args.post_treatment,
        dim=args.dim,
        seed=args.seed,
    )
    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")

    error_all, error_null = evaluate_baselines(data)
    print(f"error_all: {error_all}")

    kiv = Kernelized(args.dim + 1, args.n_env, args.linear_time)
    subset_kiv = kiv.compute_subset(X, Y, T)
    error_kiv = evaluate_subset(subset_kiv, data)

    print(f"Subset KIV: {subset_kiv}")

    subset_kiv_gumbel = tune_kiv_gumbel(args, X, Y, T)
    print(f"Subset KIV gumbel: {subset_kiv_gumbel}")
    error_kiv_gumbel = evaluate_subset(subset_kiv_gumbel, data)

    error_irm = evaluate_irm(data)

    print(
        f"MAE (null):{error_null:.3f}\nMAE (all):{error_all:.3f}\nMAE(kiv):{error_kiv:.3f}\nMAE (kiv_gumbel):{error_kiv_gumbel:.3f}\nMAE (irm):{error_irm:.3f}"
    )


if __name__ == "__main__":
    main()
