import numpy as np
import matplotlib.pyplot as plt
from tools import *
from matplotlib.backends.backend_pdf import PdfPages
import os
from copy import deepcopy
import cvxpy as cp
import causaldag as cd
from causaldag import (
    partial_correlation_suffstat,
    partial_correlation_test,
    MemoizedCI_Tester,
)
import networkx as nx
import random
import argparse

import torch


def visualize_graph(G, title, save_path=None, seed=2):
    g = G.to_nx()
    pos = nx.spring_layout(g, seed=seed)
    plt.figure()
    nx.draw(g, with_labels=True, pos=pos, node_color="lightblue", node_size=100)
    plt.title(title)
    if save_path is not None:
        plt.savefig(save_path)


def find_T_Y(dag):
    for node in dag.nodes:
        if dag.outdegree_of(node) >= 1:
            T = node
            Y = list(dag.children_of(node))[0]
            return T, Y
    return None, None


def generate_data(
    dag, B, sigma, n=1000, t_ind=0, y_ind=1, shift=None, observed_nodes=None
):
    data = {}
    for i in dag.topological_sort():
        parents = list(dag.parents_of(i))
        if i == t_ind or i == y_ind:
            data[i] = np.random.normal(0, sigma[i], n)
        else:
            data[i] = np.random.normal(shift[i], sigma[i] * shift[i] ** 2, n)
        for j in parents:
            data[i] += B[j, i] * data[j]
        if i == t_ind:
            p = torch.sigmoid(torch.Tensor(data[i]))
            T = torch.bernoulli(p)
            data[i] = T.numpy()

    data_list = []
    for i in range(len(data)):
        data_list.append(data[i])
    data = np.array(data_list).T
    return data


def check_mediators(dag, T, Y):
    gg = dag.copy()
    gg.remove_arc(T, Y)
    if Y in gg.descendants_of(T):
        print("There is an indirect path from T to Y")
        return True
    else:
        print("No mediators")
        print("Descendants of T w.o. Y: ", gg.descendants_of(T))
        return False


def check_collider(dag, T, Y):
    # output True if there is a collider on the path from T to Y
    if dag.children_of(T) & dag.children_of(Y) != set():
        print("Collider present")
        return True
    print("No collider")
    return False


def check_confounder(dag, T, Y):
    # output True if there is a collider on the path from T to Y
    if dag.parents_of(T) & dag.parents_of(Y) != set():
        print("Confounder present")
        return True
    print("No Confounder")
    return False


def main(d: int, n: int, connectivity: float, num_envs: int, seed=2, output_path=None):
    # %%
    print("Generating random graph")
    random.seed(seed)
    np.random.seed(seed)

    T_or_Y_observed = random.choice(
        [0, 1]
    )  # if 0, parents of T are fully observed. if 1, parents of Y are fully observed
    mediators_flag = True
    confounder_flag = False
    unobserved_flag = False

    # %%
    while (mediators_flag) or (not confounder_flag) or (not unobserved_flag):
        nodes = set(range(d))
        dag = cd.rand.directed_erdos(d, connectivity)
        gaussdag = cd.rand.rand_weights(dag)
        B = gaussdag.weight_mat
        sigma = (
            gaussdag.variances
        )  # vector of variances of epsilons (independent exogenous noises of nodes)
        Sigma = gaussdag.covariance

        T, Y = find_T_Y(dag)
        # print parents, children of T and Y
        mediators_flag = check_mediators(dag, T, Y)
        confounder_flag = check_confounder(dag, T, Y)

        if args.invariance == "Y":
            forbidden_nodes = set(dag.parents_of(Y)) | {Y}
            unobserved_nodes = list(
                (dag.ancestors_of(Y) | dag.ancestors_of(T)) - forbidden_nodes
            )
        elif args.invariance == "T":
            forbidden_nodes = set(dag.parents_of(T)) | {T} | {Y}
            unobserved_nodes = list(
                (dag.ancestors_of(Y) | dag.ancestors_of(T)) - forbidden_nodes
            )
        else:
            unobserved_nodes = [0]

        unobserved_flag = False if len(unobserved_nodes) == 0 else True

    visualize_graph(dag, "Simulated DAG", save_path=output_path + "_dag")
    # %%
    all_nodes = set(range(d))
    # divide nodes into observed and confounders. make sure that either the parents of T or the parents of Y are observed
    if args.invariance == "Y":
        forbidden_nodes = set(dag.parents_of(Y)) | {Y}
        unobserved_nodes = list(
            (dag.ancestors_of(Y) | dag.ancestors_of(T)) - forbidden_nodes
        )

    elif args.invariance == "T":
        forbidden_nodes = set(dag.parents_of(T)) | {T} | {Y}
        unobserved_nodes = list(
            (dag.ancestors_of(Y) | dag.ancestors_of(T)) - forbidden_nodes
        )

    else:
        forbidden_nodes = set(dag.parents_of(Y)) | set(dag.parents_of(T)) | {Y} | {T}
        unobserved_nodes = list(
            (dag.descendants_of(Y) | dag.descendants_of(T)) - forbidden_nodes
        )

    # all_parents =   dag.ancestors_of(Y) | set(dag.parents_of(Y)) - set_parents
    # unobserved_nodes = random.sample(sample_nodes, int(len(sample_nodes)))
    observed_nodes = all_nodes - set(unobserved_nodes)
    full_observed_data = list()
    ate = B[T, Y]
    new_T_ind = list(observed_nodes).index(T)
    new_Y_ind = list(observed_nodes).index(Y)
    for i in range(num_envs):
        shift = np.random.uniform(size=sigma.shape[0], low=-1.0, high=1.0)
        # shift = np.random.normal(size=sigma.shape[0], scale=1)
        data = generate_data(dag, B, sigma, n=n, t_ind=T, y_ind=Y, shift=shift)
        observed_data = data[:, list(observed_nodes)]

        shift = np.random.uniform(low=-1.0, high=1.0)

        if args.invariance != "YT":
            collider_column = (
                observed_data[:, new_Y_ind]
                + observed_data[:, new_T_ind]
                + np.random.normal(loc=shift, scale=shift**2)
            )
            observed_data = np.column_stack((observed_data, collider_column))

        full_observed_data.append(observed_data)

    if args.invariance != "YT":
        bad_controls = list(dag.children_of(Y))
        for i in range(len(bad_controls)):
            bad_controls[i] = list(observed_nodes).index(bad_controls[i])
    else:
        bad_controls = []

    good_controls = list(dag.parents_of(T) & dag.parents_of(Y))
    for i in range(len(good_controls)):
        good_controls[i] = list(observed_nodes).index(good_controls[i])

    # save full_data list, new_T_ind, new_Y_ind in a .csv file
    if output_path is not None:
        np.savez(
            output_path,
            observed_data=full_observed_data,
            new_T_ind=new_T_ind,
            new_Y_ind=new_Y_ind,
            ate=ate,
            bad_controls=list(bad_controls),
            good_controls=good_controls,
        )
    else:
        # check if "data" folder exists
        if not os.path.exists("data"):
            os.makedirs("data")
        np.savez(
            f"data/random_graphs.npz",
            observed_data=full_observed_data,
            new_T_ind=new_T_ind,
            new_Y_ind=new_Y_ind,
            ate=ate,
            bad_controls=bad_controls,
            good_controls=good_controls,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Produce multi-environment data from a random graph with prespecified parameters"
    )
    parser.add_argument("--d", type=int, default=20, help="Dimension of the graph")
    parser.add_argument(
        "--n", type=int, default=20000, help="Number of samples for each environment"
    )
    parser.add_argument(
        "--connectivity", type=float, default=0.5, help="Connectivity of the graph"
    )
    parser.add_argument(
        "--invariance", type=str, default="YT", help="Unobserved parents"
    )

    parser.add_argument(
        "--num_envs", type=int, default=5, help="Number of environments"
    )
    parser.add_argument("--seed", type=int, required=False, help="Random seed")
    parser.add_argument(
        "--output_path", type=str, default="test", help="Path to save the data"
    )

    args = parser.parse_args()
    main(args.d, args.n, args.connectivity, args.num_envs, args.seed, args.output_path)
