import torch

def gaussian_kernel(xi, xj, sigma):
    """Gaussian kernel function."""
    dist_sq = torch.cdist(xi, xj, p=2) ** 2
    return torch.exp(-dist_sq / (2 * sigma ** 2))


def compute_linear_stat(x, psi):
    # Number of pairs to consider is the smaller of the two: len(x[::2]) and len(x[1::2])
    num_pairs = min(len(x[::2]), len(x[1::2]))

    # Slice the first 'num_pairs' entries to ensure pairs are of the same length
    x1 = x[::2][:num_pairs]
    x2 = x[1::2][:num_pairs]

    # Compute pairwise distance
    dist = torch.norm(x1 - x2, dim=1)

    # Compute sigma as the median of the distances
    sigma = dist.median()

    # Compute kernel using the Gaussian formula
    kernel = torch.exp(-dist ** 2 / (2 * sigma ** 2))

    # Compute h using the first 'num_pairs' entries of psi to match x1 and x2
    h = psi[::2][:num_pairs] * psi[1::2][:num_pairs] * kernel

    # Calculate the statistic
    stat = h.mean().abs() / (h.std() + 1e-8)  # Adding epsilon to avoid division by zero

    return stat


def construct_kernel_matrix(x):
    n = x.shape[0]

    # Expand dimensions to use broadcasting for pairwise operations
    x_i = x.unsqueeze(1)  # Shape: (n, 1, d)
    x_j = x.unsqueeze(0)  # Shape: (1, n, d)

    # Pairwise kernel application using broadcasting
    kernel_param = torch.cdist(x,x).flatten().median()

    kernel_matrix = gaussian_kernel(x_i, x_j, kernel_param)

    return kernel_matrix.reshape(n,n)



def construct_cross_kernel_matrix(x):
    n = x.shape[0]
    m = n // 2

    # Create meshgrid for indices
    idx_i = torch.arange(m)
    idx_j = torch.arange(m, n)

    # Flatten and index into x to get the pairs
    x_i = x[idx_i]
    x_j = x[idx_j]

    # Apply kernel function to all pairs in a vectorized manner
    kernel_param = torch.cdist(x,x).flatten().median()
    kernel_matrix = gaussian_kernel(x_i, x_j, kernel_param)

    return kernel_matrix

def compute_cross_statistic(psi, K):
    m1 = K.shape[0]
    m2 = K.shape[1]
    
    # Create meshgrid for indices
    idx_i, idx_j = torch.meshgrid(torch.arange(m1), torch.arange(m1, m1 + m2), indexing='ij')

    # Flatten and index into psi to get the pairs
    psi_i = psi[idx_i.flatten()]
    psi_j = psi[idx_j.flatten()]

    # Compute hhh
    hhh = psi_i * psi_j * K.flatten()

    # Compute fhs
    hhh = hhh.reshape(m1, m2)
    fhs = hhh.sum(dim=1) / m2

    # Compute U
    U = fhs.mean() / fhs.std()

    # Compute loss
    loss = U

    return torch.abs(loss)


def compute_u_statistic(psi, K):
    n = K.shape[0]
    
    # Create meshgrid for indices
    psi = psi.view(-1, 1)  # Reshape psi to (m1 + m2, 1)

    psi_outer = psi @ psi.T  # This gives a matrix of shape (m1 + m2, m1 + m2)

    hhh = psi_outer * K  * (1-torch.eye(n))

    fhs = hhh.mean() 


    # Compute loss
    loss = fhs

    return torch.abs(loss)
    