import random

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

np.random.seed(0)
torch.manual_seed(0)
random.seed(0)


def mean_nll(y_logit, y):
    return nn.functional.binary_cross_entropy_with_logits(y_logit, y.float())


class IRM_MLP(nn.Module):
    def __init__(self, environments, args):
        super(IRM_MLP, self).__init__()
        best_reg = 0
        best_err = 1e6
        self.environments = environments
        self.args = args

        dim_x = environments[-1][0].size(1)
        hidden_dim = 200
        hypo_dim = 100

        # Define layers for the model
        self.lin1 = nn.Linear(dim_x, hidden_dim).cuda()
        self.lin1_1 = nn.Linear(hidden_dim, hypo_dim).cuda()

        # Define treatment heads (for treated and untreated)
        if args["net"] == "tarnet":
            self.lin1_3 = nn.Linear(dim_x, 1).cuda()
        else:
            self.lin1_3 = nn.Linear(hypo_dim, 1).cuda()

        self.lin2_0 = nn.Linear(hypo_dim, hypo_dim).cuda()
        self.lin2_1 = nn.Linear(hypo_dim, hypo_dim).cuda()

        self.lin4_0 = nn.Linear(hypo_dim, 1).cuda()
        self.lin4_1 = nn.Linear(hypo_dim, 1).cuda()

        for lin in [
            self.lin1,
            self.lin1_1,
            self.lin2_0,
            self.lin2_1,
            self.lin1_3,
            self.lin4_0,
            self.lin4_1,
        ]:
            nn.init.xavier_uniform_(lin.weight)
            nn.init.zeros_(lin.bias)

        self.train_model(environments, args)

    def forward(self, input):
        x = F.relu(self.lin1(input))
        x = F.relu(self.lin1_1(x))

        t_pred = self.lin1_3(x) if self.args["net"] != "tarnet" else self.lin1_3(input)

        h1 = F.relu(self.lin2_1(x))
        h0 = F.relu(self.lin2_0(x))

        y1 = self.lin4_1(h1)
        y0 = self.lin4_0(h0)

        return y0, y1, t_pred

    def train_model(self, environments, args):

        optimizer_adam = torch.optim.Adam(self.parameters(), lr=1e-3)
        optimizer_sgd = torch.optim.SGD(self.parameters(), lr=1e-7, momentum=0.9)
        loss_fn = nn.MSELoss()

        for iteration in range(200):
            penalty = 0
            e = 0
            e_t = 0
            for x_e, t_e, p_e, y_e in environments:
                x_e, t_e, y_e = x_e.cuda(), t_e.cuda(), y_e.cuda()
                # Forward pass
                y0_pred, y1_pred, t_pred = self.forward(x_e)

                # Select prediction based on treatment assignment
                y_pred = t_e * y1_pred + (1 - t_e) * y0_pred
                # Compute the error and penalty for IRM
                error_y = loss_fn(y_pred, y_e)
                error_t = mean_nll(t_pred, t_e)
                scale = torch.tensor(1.0).cuda().requires_grad_()
                loss_penalty = loss_fn(y_pred * scale, y_e)
                penalty += (
                    torch.autograd.grad(loss_penalty, [scale], create_graph=True)[0]
                    .pow(2)
                    .sum()
                )
                e += error_y
                e_t += error_t

            e /= len(environments)
            e_t /= len(environments)

            weight_norm = torch.tensor(0.0).cuda()
            for w in self.parameters():
                weight_norm += w.norm().pow(2)

            e += 0.001 * weight_norm
            penalty_weight = 100.0 if iteration >= 100 else 1
            e += penalty_weight * penalty
            e += e_t

            if penalty_weight > 1.0:
                # Rescale the entire loss to keep gradients in a reasonable range
                e /= penalty_weight

            if iteration < 501:
                optimizer_adam.zero_grad()
                e.backward()
                optimizer_adam.step()
            # train with sgd after the performance is stabilized with adam.
            else:
                optimizer_sgd.zero_grad()
                e.backward()
                optimizer_sgd.step()

    def ate(self, n_env):
        ates = []
        for e in range(n_env):
            X = self.environments[e][0].cuda()  
            # Apply the forward method for ATE calculation
            y0_pred, y1_pred, _ = self.forward(X)
            ite = y1_pred - y0_pred
            ates.append(ite.mean())
        return ates
