import concurrent
import time
from functools import partial
from itertools import chain, combinations

import numpy as np
import pandas as pd
import plotly.express as px
import torch
import wandb
from sklearn.linear_model import LinearRegression, LogisticRegression
from xgboost import XGBClassifier, XGBRegressor

from kernel_utils import (
    compute_cross_statistic,
    compute_linear_stat,
    construct_cross_kernel_matrix,
)
from utils import XGB_PARAMS


def generate_subsets(data):
    return chain.from_iterable(combinations(data, r) for r in range(len(data) + 1))


class Kernelized:
    def __init__(
        self, input_dim: int, n_env: int, linear_time=False, use_xgboost=False
    ):
        super(Kernelized, self).__init__()
        self.input_dim = input_dim
        self.n_env = n_env
        self.linear_time = linear_time
        self.use_xgboost = use_xgboost

    def compute_stat(self, x, psi):
        if self.linear_time:
            mmd = compute_linear_stat(x, psi)
        else:
            K = construct_cross_kernel_matrix(x)
            mmd = compute_cross_statistic(psi, K)
        return mmd

    def compute_subset(self, X, Y, T):
        start_time = time.time()
        X = torch.from_numpy(X)
        T = torch.from_numpy(T)
        Y = torch.from_numpy(Y)
        tmp = [i for i in range(self.input_dim)]
        subsets = list(generate_subsets(tmp))
        subsets.remove(())
        partial_loss = partial(self.compute_loss, X=X, Y=Y, T=T)
        with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
            results = list(executor.map(partial_loss, subsets))
        best_subset = min(results, key=lambda x: x[0])[1]

        print(f"Best subset_kiv: {best_subset}")
        subset_sorted = sorted(results, key=lambda x: x[0])
        loss_values = [item[0] for item in subset_sorted]
        subsets = [str(item[1]) for item in subset_sorted]
        df = pd.DataFrame(
            {"Subset": [str(i) for i in subsets], "Loss Value": loss_values}
        )

        # Create a scatter plot using plotly
        fig = px.scatter(df, x="Subset", y="Loss Value")
        wandb.log({"my_chart": fig})

        total_time = time.time() - start_time
        print(f"Total time for compute_subset: {total_time:.5f} seconds")

        return best_subset

    def compute_loss(self, subset, X, Y, T):
        print(f"Computing loss for subset: {subset}")

        X_pool = X.reshape(-1, X.shape[2])[:, subset]
        T_pool = T.reshape(-1)
        Y_pool = Y.reshape(-1)

        if self.use_xgboost:
            print("Using XGBoost for model training.")
            mu_t = XGBClassifier(**XGB_PARAMS)
            mu_y1 = XGBRegressor(**XGB_PARAMS)
            mu_y0 = XGBRegressor(**XGB_PARAMS)
        else:
            print("Using Logistic and Linear Regression for model training.")
            mu_t = LogisticRegression(max_iter=500)
            mu_y1 = LinearRegression()
            mu_y0 = LinearRegression()

        mu_t.fit(X_pool, T_pool)
        mu_y1.fit(X_pool[T_pool == 1], Y_pool[T_pool == 1])
        mu_y0.fit(X_pool[T_pool == 0], Y_pool[T_pool == 0])

        print(f"Model training completed for subset: {subset}")

        # Initialize loss lists
        loss_t = []
        loss_y = []

        for e in range(X.shape[0]):
            X_e = X[e, :, subset]
            T_e = T[e, :]
            Y_e = Y[e, :]

            # Compute T-invariance loss
            psi = T_e - mu_t.predict_proba(X_e)[:, 1]
            mmd_t = self.compute_stat(X_e, psi)
            loss_t.append(mmd_t)

            # Compute Y-invariance loss for T=1 and T=0
            psi = Y_e[T_e == 1] - mu_y1.predict(X_e[T_e == 1])
            mmd_y1 = self.compute_stat(X_e[T_e == 1], psi)
            loss_y.append(mmd_y1)

            psi = Y_e[T_e == 0] - mu_y0.predict(X_e[T_e == 0])
            mmd_y0 = self.compute_stat(X_e[T_e == 0], psi)
            loss_y.append(mmd_y0)

        loss_t = np.asarray(loss_t).max()
        loss_y = np.asarray(loss_y).max()

        print(
            f"Loss for subset {subset}: (T-invariance: {loss_t:.5f}, Y-invariance: {loss_y:.5f})"
        )

        return min(loss_t, loss_y), subset
