import pdb
from collections import OrderedDict

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm

from kernel_utils import compute_cross_statistic, construct_cross_kernel_matrix
from models.Kernelized import Kernelized


def sample_gumbel(shape, device, eps=1e-20):
    U = torch.rand(shape, device=device)
    return -torch.log(-torch.log(U + eps) + eps)


class GumbelGate(torch.nn.Module):
    # Credits to: https://github.com/wmyw96/FAIR/tree/main

    def __init__(self, input_dim, init_offset=0, device="cpu"):
        super(GumbelGate, self).__init__()
        self.logits = torch.nn.Parameter(
            (torch.rand(input_dim, device=device)) * 1e-5 + init_offset
        )
        self.input_dim = input_dim
        self.device = device

    def generate_mask(self, temperature, shape=None):
        if shape is None:
            shape = (1, self.input_dim)
        gumbel_softmax_sample = (
            self.logits
            + sample_gumbel(self.logits.shape, self.device)
            - sample_gumbel(self.logits.shape, self.device)
        )
        mask = torch.sigmoid(gumbel_softmax_sample / temperature)
        return torch.reshape(mask, shape)


class NN_Y(nn.Module):
    def __init__(self, input_dim):
        super(NN_Y, self).__init__()
        self.fc = nn.Linear(input_dim, 1)  # Linear layer (input_dim -> 1)

    def forward(self, x):
        logits = self.fc(x)
        return logits


class NN_T(nn.Module):
    def __init__(self, input_dim):
        super(NN_T, self).__init__()
        self.fc = nn.Linear(input_dim, 1)  # Linear layer (input_dim -> 1)

    def forward(self, x):
        # Forward pass: Apply linear transformation and use sigmoid activation
        logits = self.fc(x)
        probs = torch.sigmoid(logits)  # Output probabilities in [0, 1]
        return probs


class KernelizedGumbel(nn.Module):
    def __init__(
        self,
        input_dim: int,
        n_env: int,
        init_temp: float = 1.0,
        final_temp: float = 0.001,
        anneal_iter: int = 100,
        learning_rate: float = 1e-3,
        anneal_rate: float = 0.95,
        n_epochs: int = 500,
        patience: int = 50,
        use_xgboost: bool = False,
    ):
        super(KernelizedGumbel, self).__init__()
        self.input_dim = input_dim
        self.init_temp = init_temp
        self.final_temp = final_temp
        self.anneal_iter = anneal_iter
        self.anneal_rate = anneal_rate
        self.n_epochs = n_epochs
        self.patience = patience
        self.use_xgboost = use_xgboost
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Initialize model components
        self.mu_t = NN_T(input_dim).to(self.device)
        self.mu_y1 = NN_Y(input_dim).to(self.device)
        self.mu_y0 = NN_Y(input_dim).to(self.device)
        self.set_t = GumbelGate(input_dim, device=self.device)
        self.set_y = GumbelGate(input_dim, device=self.device)
        self.n_env = n_env
        self.lr = learning_rate

    def compute_subset(self, X, Y, T):
        X = torch.from_numpy(X).float().to(self.device)
        T = torch.from_numpy(T).float().to(self.device)
        Y = torch.from_numpy(Y).float().to(self.device)

        print("KernelizedGumbel: Training T-loss...")
        mask_t = self.train_t(X, Y, T)
        print("KernelizedGumbel: Training Y-loss...")
        mask_y = self.train_y(X, Y, T)

        subset_y = torch.where(mask_y == 1.0)[1]
        subset_t = torch.where(mask_t == 1.0)[1]
        kiv = Kernelized(self.input_dim, self.n_env, use_xgboost=self.use_xgboost)

        loss_y = (
            kiv.compute_loss(subset_y.cpu(), X.cpu(), Y.cpu(), T.cpu())[0]
            if len(subset_y) > 0
            else np.inf
        )

        loss_t = (
            kiv.compute_loss(subset_t.cpu(), X.cpu(), Y.cpu(), T.cpu())[0]
            if len(subset_t) > 0
            else np.inf
        )

        subset = subset_t if loss_t < loss_y else subset_y

        return subset.cpu().numpy(), min(loss_t, loss_y)

    def compute_loss_y(self, X, Y, T, mask):
        losses = 0
        for e in range(self.n_env):
            X_e = X[e] * mask
            T_e = T[e, :]
            Y_e = Y[e, :]
            K = construct_cross_kernel_matrix(X_e[T_e == 1])
            psi = Y_e[T_e == 1] - self.mu_y1(X_e[T_e == 1]).reshape(-1)
            mmd_y1 = compute_cross_statistic(psi, K)
            losses += mmd_y1
            K = construct_cross_kernel_matrix(X_e[T_e == 0])
            psi = Y_e[T_e == 0] - self.mu_y0(X_e[T_e == 0]).reshape(-1)
            mmd_y0 = compute_cross_statistic(psi, K)
            losses += mmd_y0

        return losses.mean()

    def compute_loss_t(self, X, T, mask):
        losses = 0
        for e in range(self.n_env):
            X_e = X[e] * mask
            T_e = T[e, :]
            K = construct_cross_kernel_matrix(X_e)
            psi = T_e - self.mu_t(X_e).reshape(-1)
            mmd_t = compute_cross_statistic(psi, K)
            losses += mmd_t

        loss_t = losses.mean()
        return loss_t

    def train_t(self, X, Y, T):
        tau = self.init_temp
        optimizer_set_t = optim.Adam(self.set_t.parameters(), lr=self.lr)
        optimizer_mut = optim.SGD(self.mu_t.parameters(), lr=0.01)

        best_loss_t = float("inf")
        epochs_without_improvement = 0

        for i in tqdm(range(self.n_epochs)):
            if (i + 1) % self.anneal_iter == 0:
                tau = max(self.final_temp, tau * self.anneal_rate)

            optimizer_set_t.zero_grad()
            optimizer_mut.zero_grad()

            mask_t = self.set_t.generate_mask(tau)
            loss_t = self.compute_loss_t(X, T, mask_t)
            loss_t.backward()
            optimizer_set_t.step()
            optimizer_mut.step()

            if loss_t < best_loss_t:
                best_loss_t = loss_t
                epochs_without_improvement = 0
            else:
                epochs_without_improvement += 1

            if epochs_without_improvement >= self.patience:
                break

        return self.set_t.generate_mask(self.final_temp)

    def train_y(self, X, Y, T):
        tau = self.init_temp
        optimizer_set_y = optim.Adam(self.set_y.parameters(), lr=self.lr)
        optimizer_y1 = optim.SGD(self.mu_y1.parameters(), lr=0.01)
        optimizer_y0 = optim.SGD(self.mu_y0.parameters(), lr=0.01)

        best_loss_y = float("inf")
        epochs_without_improvement = 0

        for i in tqdm(range(self.n_epochs)):
            if (i + 1) % self.anneal_iter == 0:
                tau = max(self.final_temp, tau * self.anneal_rate)

            optimizer_set_y.zero_grad()
            optimizer_y0.zero_grad()
            optimizer_y1.zero_grad()

            mask_y = self.set_y.generate_mask(tau)
            loss_y = self.compute_loss_y(X, Y, T, mask_y)
            loss_y.backward()
            optimizer_set_y.step()
            optimizer_y1.step()
            optimizer_y0.step()

            if loss_y < best_loss_y:
                best_loss_y = loss_y
                epochs_without_improvement = 0
            else:
                epochs_without_improvement += 1

            if epochs_without_improvement >= self.patience:
                break

        return self.set_y.generate_mask(self.final_temp)
