import numpy as np
import torch
from sklearn.linear_model import LinearRegression, LogisticRegression
from xgboost import XGBClassifier, XGBRegressor

from models.IRM import IRM
from models.IRM_MLP import IRM_MLP

epsilon = 0.025

XGB_PARAMS = {
    "n_estimators": 500,
    "learning_rate": 0.03,
    "max_depth": 6,
}


def pool_data(data, dim):
    pooled = []
    for _ in data:
        pooled.append(_[dim])
    return np.asarray(pooled)


def evaluate_baselines(data, use_xgboost=False):
    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")

    n_env = len(data)
    X_pool = X.reshape(-1, X.shape[2])
    T_pool = T.reshape(-1)
    Y_pool = Y.reshape(-1)

    # Select models based on the flag
    if use_xgboost:
        mu_t_all = XGBClassifier(**XGB_PARAMS)
        mu_y1_all = XGBRegressor(**XGB_PARAMS)
        mu_y0_all = XGBRegressor(**XGB_PARAMS)
    else:
        mu_t_all = LogisticRegression()
        mu_y0_all = LinearRegression()
        mu_y1_all = LinearRegression()

    # Fit the models
    mu_t_all.fit(X_pool, T_pool)
    mu_y0_all.fit(X_pool[T_pool == 0], Y_pool[T_pool == 0])
    mu_y1_all.fit(X_pool[T_pool == 1], Y_pool[T_pool == 1])

    error_all = 0
    error_null = 0
    for e in range(n_env):
        T_e = T[e, :]
        Y_e = Y[e, :]
        ate_null = Y_e * T_e / ((T_e == 1).mean()) - Y_e * (1 - T_e) / (
            (T_e == 0).mean()
        )
        error_null += np.abs(data[e]["ATE"] - ate_null.mean())
        X_e = X[e, :, :].reshape(-1, X.shape[2])
        pi1_all = mu_t_all.predict_proba(X_e)[:, 1]
        pi1_all = np.clip(pi1_all, a_min=epsilon, a_max=1 - epsilon)
        pi0_all = 1 - pi1_all
        ate_all = (
            mu_y1_all.predict(X_e)
            - mu_y0_all.predict(X_e)
            + T_e / pi1_all * (Y_e - mu_y1_all.predict(X_e))
            - (1 - T_e) / pi0_all * (Y_e - mu_y0_all.predict(X_e))
        )
        error_all += np.abs(data[e]["ATE"] - ate_all.mean())

    error_all /= n_env
    error_null /= n_env
    return error_all, error_null


def evaluate_subset(subset, data, use_xgboost=False):
    import pdb

    if len(subset) == 0:
        return evaluate_baselines(data, use_xgboost)[1]

    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")

    n_env = len(data)

    X_pool = X[:, :, subset].reshape(-1, len(subset))
    T_pool = T.reshape(-1)
    Y_pool = Y.reshape(-1)

    # Select models based on the flag
    if use_xgboost:
        mu_t = XGBClassifier(**XGB_PARAMS)
        mu_y0 = XGBRegressor(**XGB_PARAMS)
        mu_y1 = XGBRegressor(**XGB_PARAMS)
    else:
        mu_t = LogisticRegression()
        mu_y0 = LinearRegression()
        mu_y1 = LinearRegression()

    # Fit the models
    mu_t.fit(X_pool, T_pool)
    mu_y0.fit(X_pool[T_pool == 0], Y_pool[T_pool == 0])
    mu_y1.fit(X_pool[T_pool == 1], Y_pool[T_pool == 1])

    error = 0
    for e in range(n_env):
        T_e = T[e, :]
        Y_e = Y[e, :]
        X_e = X[e, :, subset].T
        pi1 = mu_t.predict_proba(X_e)[:, 1]
        pi1 = np.clip(pi1, a_min=epsilon, a_max=1 - epsilon)
        pi0 = 1 - pi1

        hat_ate = (
            mu_y1.predict(X_e)
            - mu_y0.predict(X_e)
            + T_e / pi1 * (Y_e - mu_y1.predict(X_e))
            - (1 - T_e) / pi0 * (Y_e - mu_y0.predict(X_e))
        )
        error += np.abs(data[e]["ATE"] - hat_ate.mean())

    error /= n_env

    return error


def evaluate_irm(data):
    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")
    n_env = len(data)

    environments = []
    for e in range(n_env):
        X_e = np.asarray(X[e, :, :]).copy()
        intercept = np.ones((X_e.shape[0], 1))  # Create a column of ones
        X_e_with_intercept = np.hstack((intercept, X_e))
        T_e = np.asarray(T[e, :]).copy().reshape(-1, 1)
        Y_e = np.asarray(Y[e, :]).copy().reshape(-1, 1)
        environments.append(
            (torch.Tensor(X_e_with_intercept), torch.Tensor(T_e), 0, torch.Tensor(Y_e))
        )

    irm = IRM(environments, {})
    irm_ates = irm.ate(n_env)
    error = 0
    for e in range(n_env):
        error += np.abs(irm_ates[e].detach().numpy() - data[e]["ATE"])
    error /= n_env
    return error


def evaluate_irm_nonlinear(data):
    # import pdb
    # pdb.set_trace()
    X = pool_data(data, "X")
    T = pool_data(data, "T")
    Y = pool_data(data, "Y")
    n_env = len(data)

    environments = []
    for e in range(n_env):
        X_e = np.asarray(X[e, :, :]).copy()
        T_e = np.asarray(T[e, :]).copy().reshape(-1, 1)
        Y_e = np.asarray(Y[e, :]).copy().reshape(-1, 1)
        environments.append(
            (torch.Tensor(X_e), torch.Tensor(T_e), 0, torch.Tensor(Y_e))
        )

    irm = IRM_MLP(environments, {"net": "tarnet"})
    irm_ates = irm.ate(n_env)
    error = 0
    for e in range(n_env):
        error += np.abs(irm_ates[e].cpu().detach().numpy() - data[e]["ATE"])
    error /= n_env
    return error
